// Distributed under GNU General Public License (see license.txt for details).
//
//  Copyright (c) 2013 Tianbao Yang.
//  All Rights Reserved.
//============================================================================================
// File Name: data_set.h
// header for Datset class
//==============================================================================================


#ifndef DATASET_H_
#define DATASET_H_


#include <iostream>
#include <fstream>
#include <cstdio>
#include <cstdlib>
#include <ostream>
#include <iterator>
#include <vector>
#include <set>
#include <sstream>
#include <ctime>
#include <cmath>
#include <algorithm>
#include <type_traits>
#include "WeightVector.h"

typedef unsigned int 	uint;
typedef unsigned long	ulong;


/** Implements a dense vector
*/
class Dataset {
 public:

  /** Default Constructor.
  */
  Dataset(): num_examples(0),  dimension(0), format(NULL), task(NULL),  normalize(false), bias(0.0), readingTime(0)
  {  }

  /** Constructor. Construct a vector from an istringstream.
      @param is The input stringstream to read from.
      @param n The number of elements to read
  */
  
  Dataset(std::string fmt, std::string tsk, uint d, bool norm=false, float b=0.0): num_examples(0), dimension(d), format(fmt), task(tsk),  normalize(norm), bias(b), readingTime(0) 
  {}
  
  Dataset(std::string& data_filename, std::string fmt, std::string tsk, uint d, bool norm=false, float b=0.0): num_examples(0), dimension(d), format(fmt), task(tsk),  normalize(norm), bias(b), readingTime(0)
  {
	  ReadData(data_filename);
  }
  
  Dataset(std::string& data_filename, std::string& label_filename, std::string fmt, std::string tsk, uint d, bool norm=false, float b=0.0): num_examples(0), dimension(d), format(fmt), task(tsk),  normalize(norm), bias(b), readingTime(0)
  {
	 if(label_filename=="noLabelFile")
	    ReadData(data_filename);
	 else{
		if(dimension==0){
	       std::cerr<<"input dimension not given!"<<std::endl;
	       exit(EXIT_FAILURE);
	    } 
	    ReadBinaryData(data_filename);
	    ReadLabels(label_filename);
	 }
  }  
  
  void Append(std::string& data_filename){
	ReadData(data_filename);  
  }
  
  void Append(std::string& data_filename, std::string& label_filename){
	  if(label_filename=="noLabelFile")
	 	    ReadData(data_filename);
	 	 else{
	 		if(dimension==0){
	 	       std::cerr<<"input dimension not given!"<<std::endl;
	 	       exit(EXIT_FAILURE);
	 	    } 
	 	    ReadBinaryData(data_filename);
	 	    ReadLabels(label_filename);
	 	 }
  }

  
  void ReadData(std::string& data_filename);

  void ReadBinaryData(std::string& data_filename);
  
  void ReadLabels(std::string& label_filename);

    

  simple_sparse_vector& sX(ulong i){
      /*if(i>=num_examples){
         std::cerr<<"index exceed data size!"<<std::endl;
         exit(EXIT_FAILURE);
         }*/     
    return sparsedata[i];
  }
  
  dense_vector& dX(ulong i){
        /*if(i>=num_examples){
           std::cerr<<"index exceed data size!"<<std::endl;
           exit(EXIT_FAILURE);
           }*/     
      return densedata[i];
    }
  
  
  template<typename T>
  T y(ulong i) const{
      /*if(i>=num_examples){
         std::cerr<<"index exceed data size!"<<std::endl;
         exit(EXIT_FAILURE);
         }*/ 
      if(task=="class") return Labels[i];
      else return Targets[i];
  }
  
  
  double Xw(WeightVector& W, ulong i){
	  if(format=="dense") return W*densedata[i];
	  else                return W*sparsedata[i];
  }
  
  ulong size() const{
	  return num_examples;
  }
  
  uint dim() const{
	  return dimension; 
  }
  
  std::string& fmt(){
	  return format; 
  }
  
  long readtime() const{
	  return readingTime; 
  }
  
  
  // Print the content of a sparse instance to an output stream
  void print(std::ostream& os, ulong i)
  {
	  if(task=="class")	  os<<Labels[i]; 
	  else                os<<Targets[i];
	  
      if(format=="dense") densedata[i].print(os);
      else                sparsedata[i].print(os);
      os<<std::endl;
  }
  
 protected:  
  ulong num_examples; // number of exampels
  uint dimension; // dimension of data
  std::string format; // dense or sparse
  std::string task; // class or regre
  bool normalize;
  float bias; 
  std::vector<simple_sparse_vector> sparsedata;
  std::vector<dense_vector> densedata;  
  std::vector<int> Labels; // multi-class labels
  std::vector<double> Targets; 
  long readingTime; 
};

class trainDataset: public Dataset {
 public:

	 trainDataset(std::string fmt, std::string tsk, uint d, bool norm=false, float b=0.0): Dataset(fmt, tsk, d, norm, b), R(0.0), S(0.0)
	 { }
	 trainDataset(std::string& data_filename, std::string fmt, std::string tsk,  uint d, bool norm=false): Dataset(fmt, tsk, d, norm), R(0.0), S(0.0)
	  {
		  ReadData(data_filename);
	  }
	  
	 trainDataset(std::string& data_filename, std::string& label_filename, std::string fmt, std::string tsk, uint d, bool norm=false): Dataset(fmt, tsk, d, norm),  R(0.0), S(0.0)
	  {
		 ReadBinaryData(data_filename);
		 ReadLabels(label_filename);
	  }  
	  
	 void Append(std::string& data_filename){
		ReadData(data_filename);  
	  }
	  
	 void Append(std::string& data_filename, std::string& label_filename){
		 if(label_filename=="noLabelFile")
			 	    ReadData(data_filename);
		 else{
			 		if(dimension==0){
			 	       std::cerr<<"input dimension not given!"<<std::endl;
			 	       exit(EXIT_FAILURE);
			 	    } 
			 	    ReadBinaryData(data_filename);
			 	    ReadLabels(label_filename);
			 } 
	  }
	 
	  void ReadData(std::string& data_filename);

	  void ReadBinaryData(std::string& data_filename);
	  
	  void ReadLabels(std::string& label_filename);  
  
  
  
  std::set<int> mclass(){
	  return uLabels;
  }
  
  double mQ(ulong i){
	  /*if(i>=num_examples){
	         std::cerr<<"index exceed data size!"<<std::endl;
	         exit(EXIT_FAILURE);
	         }*/ 
	  return Q[i]; 
  }
  
  double mR(){
	  return R;
  }
  
  double mS(){
	  return S;
  }
  
  
 private:  
  double R; // maximum l2 norm square 
  double S; // l2 norm square of Targets
  std::vector<double> Q; //save square norm of data
  std::set<int> uLabels; // unique labels
};

#endif /*DATASET_H_*/

